.drivecode_include
!src "config.inc"
!convtab pet
!cpu 6510

;XXX TODO have sector loading in ZP instead of all this?

;--------------- ZP usage ----------------
;.. = unused
;bl = blocks on list
;ms = max sectors on current track
;fs = filesize in blocks (256 byte lock)
;tr = current track
;in = number of currently loaded block
;pl = preamble bytes lo nibbles
;ph = preamble bytes hi nibbles
;sh = sector header data (gcr encoded -> 5 lo- and hinibbles)
;ds = dirsect
;wi = windex, index to be used for in order loading
;bs = blocksize, size of transferred block

;00: .. .. .. .. .. .. .. ..
;08: bl 0e 0f 07 tr 0a 0b 03 <-+
;10: ms ds 0d 05 in 00 09 01 <-| gcr2ser table with gaps being used by other variables
;18: fs 06 0c 04 bs 02 08 wi <-+
;20: 0f 07 0d 05 0b 03 09 01 <-+ bin2ser lookup table
;28: 0e 06 0c 04 0a 02 08 00 <-+
;30: halftrack code ........
;38: is ....................
;40: here ...... .. .. .. ..
;48: 0e 08 80 sh sh sh sh sh
;50: 0f 00 00 sh sh sh sh sh
;58: 07 01 10 pl pl pl pl ..
;60: fi fi fi fi fi fi .. ..		;filestruct
;68: 0a 0c c0 ph ph ph ph ..
;70: 0b 04 40 .. .. .. .. ..
;78: 03 05 50 ff ff ff ff ff <-+
;80: ff ff ff ff ff ff ff ff <-| list of wanted sectors
;88: ff ff ff ff ff ff ff ff <-+
;90: 0d 02 20 .. .. .. .. ..
;98: 05 03 30 motor + led ..
;a0: code here ....... .. ..
;a8: 00 0f f0 .. .. .. .. ..
;b0: 09 06 60 .. .. .. .. ..
;b8: 01 07 70 idle code ....
;c0: loop here .............
;c8: 06 09 90 .. .. .. .. ..
;d0: 0c 0a a0 .. .. .. .. ..
;d8: 04 0b b0 .. .. .. .. ..
;e0: .. .. .. .. .. .. .. ..
;e8: 02 0d d0 .. .. .. .. ..
;f0: 08 0e e0 .. .. .. .. ..
;f8: .. .. .. .. .. .. .. ..
;    ^  ^  ^
;    |  |  |
;    |  |  gcr dec_hi with index << 3
;    |  gcr_dec_lo with index << 3
;    gcr2ser with index << 3
;
;thus it is sufficient to inflate the gcr values in that way that they are either right or left aligned
;-----------------------------------------

.STEPPING_SPEED	= $98

.drivecode	= $0108
.bootstrap	= $0700

;VIA1		= $1800
;VIA2		= $1c00

.VIA2_LED_OFF	= $f7
.VIA2_LED_ON	= $08

.VIA2_MOTOR_OFF	= $fb
.VIA2_MOTOR_ON	= $04

.blocks_on_list	= $08		;blocks tagged on wanted list
.track		= $0c		;current track
.max_sectors	= $10		;maximum sectors on current track
.dirsect	= $11
.index		= $14		;current blockindex
.blocks		= $18		;number of blocks the file occupies
.blocksize	= $1c
!if BITFIRE_CONFIG_IN_ORDER = 1 {
.windex		= $1f
}

.file_descriptor = $60
.to_track	= .file_descriptor + 0
.sector		= .file_descriptor + 1
.ld_addr	= .file_descriptor + 2
.file_size	= .file_descriptor + 4

.temp		= .sector
.cmp1		= .sector
.cmp2		= .index
.dest		= $52		;-> 52 is always zero for free $53 will be set, sectorheader is unused during upload

.halftrack_code	= $30
.motor_code	= $9b
.idle_code	= $bb
.init_code	= $e0

.preamble_lo	= $5b		;preamble data gcr coded lonibbles
.preamble_hi	= $6b		;preamble data gcr coded hinibbles

.wanted		= $7b		;list of wanted sectors/blockindices -> $15 bytes space

.head_lo	= $4b		;sector header gcr data lonibbles
.head_hi	= $53		;sector header gcr data hinibbles

.gcr2ser	= $00		;gcr to serial-port lookuptablea -> from $08-$1f
.bin2ser	= $20		;binary to serial data

.directory	= $0500		;directory
.lonibbles	= $0600
.hinibbles	= $0700

.gcr_dec_lo	= $f8c0
.gcr_dec_hi	= $f8a0
.gcr_dec_lo_shf	= $01
.gcr_dec_hi_shf	= $02

.DIR_SECT	= 18

BUSY		= $02
BLOCK_READY	= $08
IDLE		= $00

.bootstrap_start
!pseudopc .bootstrap {
.bootstrap_run
		lda #$12
		sta $0a
		lda #.DIR_SECT
		sta $0b

		;lda #$c0
		;sta $02
		;lda $02
		;bmi *-2

		lda #$80
		sta $02
		lda $02
		bmi *-2

		lda $1c00
		and #(.VIA2_LED_OFF & .VIA2_MOTOR_OFF)
		sta $1c00

		;XXX TODO currtrack is in $18?
                sei

		ldx #<.drivecode-1
		txs
		lda #$00
-
		sta $0100,x
		dex
		bpl -

		;lda #$08
		;sta $1800

                lda #%01111010		;DDR set bits for drivenumber to 0, ATN out, CLK out and DATA out are outputs
                sta $1802

		lda #%11101110		;CA2 manual output high, CA1 low, CB2 manual output high, CB1 low (NC)
		sta $1c0c
		lda #%01000001
		sta $1c0b

		lda #$7f		;disable all interrupts
		sta $180e
		sta $1c0e

		lda #(>.drivecode_size)+1
		sta .temp

		ldy #$00
.get_block

		ldx #BLOCK_READY	;signal that no block is ready yet, asap
		stx $1800

		ldx $1800		;wait for atn
		bpl *-3

		lda #$80		;clear all lines and set bit 7 as bit counter, to allow data in and clk in to be set/cleared by host
		sta $1800
		ldx $1800
-
		cpx $1800		;did a bit arrive? (bit flip in data in, atn is dropped in same go in first bit)
		beq *-3
		ldx $1800		;load register
		cpx #$04		;push bit into carry
		ror			;shift in
		bcc -			;do until our counter bit ends up in carry
.block = * + 1
                sta .drivecode,y

                iny
                bne .get_block

                inc .block+1
                dec .temp
                bne .get_block

		lda #BUSY
		sta $1800
		jmp .drivecode_setup
}
.bootstrap_end
.bootstrap_size = .bootstrap_end - .bootstrap_start

.drivecode_start
!pseudopc .drivecode {
.drivecode_
.read_sector
		jsr .wait_sync_mark
					;11111 222	header type
		cpx #$52
		bne .read_sector
		;lda $1c01		;22333334	type  / checksum
		asr #%00111111
		sta .head_hi+1		;store @ pos 1						;7

.chksum
		clv
		bvc *

		lax $1c01		;44445555	checksum / sector
		ror
		lsr
		lsr
		lsr
		sta .head_lo+1
		clv
					;15 cycles
		bvc *

		txa
		ldx $1c01		;56666677	sector / track
		cpx #$80
		rol
		and #%00011111
		sta .head_hi+0		;store @ pos 0 instead of pos 1 to reuse value later on
		clv
					;15 cycles
		bvc *

		txa
		asr #%01111111		;bit 4 of 7 in carry
		sta .head_lo+0		;1 bit of 7 left
		lax $1c01		;77788888	track
		ror			;-> 77778888
		lsr .head_lo+0		;fetch msb of 7

		;clv
		;bvc *

		ror			;-> 77777888
		lsr
		lsr
		lsr
		sta .head_hi+2		;-> ---77777
		lda #%00011111
		sax .head_lo+2		;-> ---88888
					;read after 40 cycles
		ldx #$07
		lda $1c01		;11111222	id2
		clv

		bvc *

		sax .head_lo+3
		lsr
		lsr
		lsr
		sta .head_hi+3		;---11111

		lax $1c01		;22333334	id2/id1
		asl
		rol .head_lo+3
		asl
		rol .head_lo+3
		txa
		asr #%00111111
					;read on cycle 40
		sta .head_hi+4
		lda $1c01		;44445555	id2
!if >* != >.chksum { !error "header-checksum not in one page: ", .chksum, " - ", * }
		ror
		lsr
		lsr
		lsr
		sta .head_lo+4

		ldx #$04
-
		ldy .head_hi,x		;decode sector header
		lda .gcr_dec_hi,y
		ldy .head_lo,x
		ora .gcr_dec_lo,y
		sta .head_lo,x
		dex
		bpl -
		tax			;sector number

		eor .head_lo + 1	;build header checksum A = .head_lo + 0
		eor .head_lo + 2
		eor .head_lo + 3
		eor .head_lo + 4
		bne .back1		;urgh, bad checksum, find another sector

		lda .head_lo + 2
		cmp .track		;check track, as we could easily read from wrong track during settling time of stepper (happend so in the wild, what a bother)
		bne .back2

		ldy .wanted,x		;sector on list?
!if BITFIRE_CONFIG_IN_ORDER = 1 {
		cpy .windex		;next sector? nope?
		bne .back
}
		iny			;$ff + 1 = 0?
		;yes, read, we should still have ~20 cycles overhead here until next sync arrives
		bne .wait_sector_data
.back
.back1
.back2
.back3
		jmp .read_sector

.wait_sector_data
		jsr .wait_sync_mark
		cpx #$55
		bne .back
		asr #%00111111
		clv
		bvc *
		ldy #$00		;waste 6 cycles to have more or less same entry point into loop
		nop
		nop

;speedzone1	1111111111111111111111111122222222222222222222222222333333333333333333333333334444444444444444444444444455555555555555555555555555
;speedzone2	11111111111111111111111111112222222222222222222222222222333333333333333333333333333344444444444444444444444444445555555555555555555555555555
;speedzone3     111111111111111111111111111111222222222222222222222222222222333333333333333333333333333333444444444444444444444444444444555555555555555555555555555555
;speedzone4     1111111111111111111111111111111122222222222222222222222222222222333333333333333333333333333333334444444444444444444444444444444455555555555555555555555555555555
;               ---1111111111111111111----      ---2222222222222----

;               0         1         2         3         4         5
;               0         0         0         0         0         0
;							^
;							|
;					best read second byte around cycle 40

;data valid at cycle:
;zone1/zone2/zone3/zone4
; 0   / 0   / 0   / 0
; 26  / 28  / 30  / 32
; 52  / 56  / 60  / 64
; 78  / 84  / 90  / 96
; 104 / 112 / 120  / 128

.load_sector_data
		sta .hinibbles,y	;---33333 4
		lax $1c01		;44445555				;read @ 16 / 16 / 16 / 16
		arr #$f0		;44444---				;/!\ ATTENTION, arr potentially fucks the v flag status!
		sta .lonibbles,y	;44444---
		txa			;44445555
		and #%00001111		;----5555
		iny
		sty .lindex+1		;avoid index on lsr

.smc1		bne *			;BRA, y is never 0, could also use bvs, but arr possibly influences V flag /o\
		nop			;so waste either 3,5,7,9 cycles depending on speedzone
		nop
		nop

		ldx $1c01		;56666677				;read @ 40 / 42 / 44 / 46
		cpx #$80		;copy 5 to carry
		rol
		sta .hinibbles,y
		txa			;56666677
		asr #%01111111		;--666667
		sta .lonibbles,y
		iny

.smc2		bne *
		nop
		nop
		nop

		lax $1c01		;77788888				;read @ 67 / 71 / 75 / 79
		and #%00011111
		sta .lonibbles,y	;lda #$0f sax .lonibbles,y if ZP
		txa			;77788888 is fetched again
		ror
.lindex		lsr .lonibbles		;shift out last 7 -> ---66666
		arr #$f0		;77777---
		sta .hinibbles,y
		lda $1c01		;11111222				;read @ 95 / 99 / 103 / 107
		clv								;now finally clear v flag, as we are past the last arr operation
		iny
		lsr

		bvc *

		sta .hinibbles,y	;-1111122 2
		lda $1c01		;22333334
		clv
		ror			;22233333 4
		sta .lonibbles,y	;store for later processing		;120 cycles min

		bvc *

		and #%00011111		;---33333 4
		iny
		bne .load_sector_data
!if >* != >.load_sector_data { !error "gcr-decoding not in one page: ", .load_sector_data, " - ", * }

		tay
		lda $1c01
		arr #$f0
		tax
		lda .gcr_dec_hi,y	;XXX TODO would also work with gcr2ser table but would reduce reliability
		ora .gcr_dec_lo_shf,x
		tay			;checksum

		;checksum/inflate whole sector only if needed
		;either check $100 bytes or just sector len & $fc + 4 bytes depending if a partial (last) or full sector is laoded
		ldx .head_lo + 0
		lda .wanted,x
		ldx #$ff		;blocksize full sector
		cmp .file_size + 1	;as we count up in $100 blocks, if index equals highbyte of size, we have loaded last block
		bne +
		ldx .file_size		;block size last sector (lowbyte filesize)
+
		;remember for later use
		stx .blocksize

		lda #$fc
		sbx #$00		;only eor as much nibbles as necessary (think of partly filled blocks), saves us some time on last block

		;finish gcr inflation and checksum the data
		;XXX TODO place code on zp and use smc instead of tay/tya.
		;place lo/hinibbles+3 on zp?
-
		lda .lonibbles + 3,x	;4
		lsr .hinibbles + 3,x	;7
		ror			;2
		lsr .hinibbles + 3,x	;7
		arr #$f0		;2
		sta .lonibbles + 3,x	;5

		tya
		ldy .hinibbles + 0,x
		eor .gcr_dec_hi,y
		ldy .lonibbles + 0,x
		eor .gcr_dec_lo_shf,y
		ldy .hinibbles + 1,x
		eor .gcr_dec_hi,y
		ldy .lonibbles + 1,x
		eor .gcr_dec_lo,y
		ldy .hinibbles + 2,x
		eor .gcr_dec_hi_shf,y
		ldy .lonibbles + 2,x
		eor .gcr_dec_lo,y
		ldy .hinibbles + 3,x
		eor .gcr_dec_hi,y
		ldy .lonibbles + 3,x
		eor .gcr_dec_lo_shf,y
		tay

		txa
		sbx #$04
		bcs -

		tya
		bne +			;checksum okay?

		dey			;y = $ff
		ldx .head_lo + 0
		lda .wanted,x		;fetch blockindex
		sty .wanted,x		;clear entry in wanted list

!if BITFIRE_CONFIG_IN_ORDER = 1 {
		inc .windex
}
		rts			;done
+
		jmp .read_sector

.send_block
		ldy #$00		;start with 0, we will send 2 or 4 bytes of blockinfo as preamble
		jsr .wait_ack_		;wait for response of host

		;on atn going low we have alraedy first data on bus with the upcoming code
.preloop
		ldx .preamble_lo,y
		lda .bin2ser,x
		bit $1800
		bmi *-3
		sta $1800		;send first bits on end of atn strobe -> 4th sta on c64 side

		asl
		ora #$10
		ldx .preamble_hi,y	;hinibbles are already shifted into right position
		bit $1800
		bpl *-3
		sta $1800		;second -> first sta $dd02 on c64 side

		lda .bin2ser,x
		iny
		bit $1800
		bmi *-3
		sta $1800		;third -> second sta $dd02 on c64 side

		asl
		ora #$10
		bit $1800
		bpl *-3
		sta $1800		;last -> third sta $dd02 on c64 side
.pre_cmp = * + 1
		cpy #$00
		bcc .preloop
!if >* != >.preloop { !error "preloop not in one page" }

		ldy .blocksize
		ldx .lonibbles,y
.sendloop				;send the data block
					;this loop is much tighter then the preamble, no cycles to be wasted
		lda .gcr2ser,x
		bit $1800
		bmi *-3
		sta $1800		;read by first lda $dd00 on c64 side

		asl
		ora #$10
		ldx .hinibbles,y
		bit $1800
		bpl *-3
		sta $1800		;read by second lda $dd00 on c64 side

		lda .gcr2ser,x
		dey
		bit $1800
		bmi *-3
		sta $1800		;read by third lda $dd00 on c64 side

		asl
		ora #$10
		ldx .lonibbles,y
		bit $1800
		bpl *-3
		sta $1800		;read by fourth lda $dd00 on c64 side
		cpy #$ff
		bcc .sendloop

		;XXX carry is set here, always, might be useful somewhen
		lda #BUSY
		bit $1800
		bmi *-3
		sta $1800

		bpl .reset_pre_cmp 	;BRA reset size for preamble

!if >* != >.sendloop { !error "sendloop not in one page" }

.preamble_size
		;clc			;cleared by read sector
		adc .ld_addr+1
		jsr .add_preamble
		ldy .blocksize		;set up num of bytes to be transferred
		iny
		tya

.add_preamble
		ldy .pre_cmp
		ldx #$0f
		sax .preamble_lo,y	;mask out lower 4 bits
		lsr
		lsr
		lsr
		lsr
		sta .preamble_hi,y	;upper 4 bits
.reset_pre_cmp
		iny
		sty .pre_cmp
		rts

.wait_sync_mark
		lda $1c00		;wait for end of sync
		bpl *-3
		lda $1c00		;wait for end of sync
		bmi *-3
		lda $1c01		;sync mark
		clv
		bvc *
		ldx $1c01		;11111222	must be $55/$52
		clv
		bvc *
		lda $1c01		;22333334
		rts
!if >* != >.wait_sync_mark { !error "wait_sync_mark not in one page: ", .wait_sync_mark, " - ", * }

.idle_
		;wait for response of host
		jsr .wait_ack_
.idle
		jsr .idle_code
		ldx #BUSY		;signal that no block is ready yet and that we are loading, asap
		stx $1800

		;load file, file number is in A
.load_file
		cmp #$f0
		bcc .load_
.turn_disc

		jsr .read_dir		;fetch first dir sector
		cmp .directory + $ff	;compare side info
		bne .turn_disc		;nope, not the requested side
		beq .idle_
.load_
		tay			;transfer to y to recheck value
		bpl *+5
		jmp .upload

		ldx #.DIR_SECT+1
		sec
-
		dex
		sbc #42
		bcs -
+
		adc #42
		cpx .dirsect		;dirsect changed?
		beq +			;nope, advance

		jsr .read_dir_
+
		sta .temp
					;file number * 6 to get proper index -> (num * 2 + num) * 2
		asl			;*2
		;clc			;can be omitted, not more than 42 files are allowed per side, so carry is always cleared after asl
		adc .temp		;+1 -> * 3
		asl			;*2 -> * 6
		tay

		jsr .motor_on		;turn on led and motor, if not on already

		ldx #$00		;reset block index
		stx .index
!if BITFIRE_CONFIG_IN_ORDER = 1 {
		stx .windex
}

-					;copy over direntry to ZP
		lda .directory,y
		sta .file_descriptor,x
		iny
		inx
		cpx #$06
		bne -
					;a = file_size highbyte, carry set
		adc #$00		;need to add one, as also the last partial block counts as a full block. File size is one too less, so exceptions like $xx00 in length are no problem
		sta .blocks
					;send load-address with first transferred block
		lda .ld_addr+1
		jsr .add_preamble
		lda .ld_addr
		jsr .add_preamble

.load_track
		lda .blocks
		beq .idle		;check if all blocks of file are loaded

		jsr .seek		;sets max_sectors and bitrate depending on track
		stx .blocks_on_list	;clear block list counter (jsr .seek returns with x = 0)
--					;now find corresponding offset for first sector by subtracting until remainder (== offset) remains
		lax .sector		;load startsector# and keep a copy in X
		sec
-
		sta .sector		;keep old value as new offset
		sbc #BITFIRE_CONFIG_INTERLEAVE;subtract some #sectors
		bcs -			;until underflow

		;we have now the right offset to start at
		;A is desired sector again, move over to X
-
		lda .index		;get index
		sta .wanted,x		;write index into wantedlist
		inc .index		;advance index (global num of blocks)
		inc .blocks_on_list	;count number of blocks in list (per track num of blocks)
		dec .blocks		;all blocks loaded?
		beq .load_wanted_blocks	;yep, end
		txa			;go for next block
		sbx #-BITFIRE_CONFIG_INTERLEAVE
		cpx .max_sectors	;wrap around?
		bcc -
		lda #BITFIRE_CONFIG_INTERLEAVE
		isc .sector
		bne --

.load_wanted_blocks			;read and transfer all blocks on wishlist
		jsr .read_sector	;returns with current blockindex in A
		jsr .preamble_size	;add blockindex and size
		;XXX TODO put send_block here
		jsr .send_block		;exists with y = 0 and carry set
		dec .blocks_on_list	;last block on wishlist?
		bne .load_wanted_blocks

.track_finished
		sty .sector		;set start pos for next track, y is always 0 after .send_block
-					;now adjust to_track and take care of skipping track 18
		lda #18
		;sec			;set by send_loop and also set if beq
		isc .to_track
		beq -			;skip dirtrack however
		bne .load_track		;BRA

.wait_ack_
		lda #BLOCK_READY | BUSY	;signal block ready/turn disc okay
		sta $1800

		bit $1800
		bpl *-3
		bit $1800
		bmi *-3
		bit $1800
		bpl *-3
		rts
.seek
.set_bitrate				;sets bitrate for desired track and also sets value for max. sectors on track
		ldy .to_track

		ldx #$11
		lda $1c00
		and #$9f
		cpy #31
		bcs +
		inx
		adc #$20
		cpy #25
		bcs +
		inx
		adc #$20
		cpy #18
		bcs +
		inx
		inx
		adc #$20
+
		stx .max_sectors
		sta $1c00

		asr #$60
		lsr
		lsr
		lsr
		lsr			;0/2/3/6
		sta .smc1+1		;adapt gcr decode to speedzone by modifying branches
		sta .smc2+1

		tya
		sec
		sbc .track		;how many tracks to go?
		sty .track		;save target track as current track

		ldy #$00
		bcs .seek_up		;up or downwards?
.seek_down
		eor #$ff		;down we go, so invert num tracks as it is negative
		adc #$01
		iny
.seek_up
		asl			;counter is twice the number of tracks (halftracks)
		tax
		beq +			;nothing to step, end
.step
		lda #.STEPPING_SPEED
		sta $1c05
.skip_wait
		jsr .halftrack
		dex
		bne .step
+
		rts

.read_dir
		ldx #.DIR_SECT
.read_dir_
		stx .dirsect
		sta .temp		;save value for later use

		lda $1c00
		;and #.VIA2_LED_OFF
		jsr .motor_on_no_led
		inc .wanted,x
		lda #$12
		sta .to_track
		sta .file_size + 1	;write any number > 0 to file_size + 1 to trigger a sector_length of 256 bytes
		jsr .seek

		jsr .read_sector	;exits with y = $ff

		;XXX TODO this is very ugly!!! :-(
		iny
-
		ldx .lonibbles,y
		cpx #$20
		bcs +
		lda .gcr_dec_lo,x
		top
+
		lda+1 .gcr_dec_lo_shf,x

		ldx .hinibbles,y
		cpx #$20
		bcs +
		ora .gcr_dec_hi,x
		top
+
		ora+1 .gcr_dec_hi_shf,x

		sta $0500,y
		dey
		bne -

		lda .temp		;restore filenum
		rts

.get_byte
		lda #$80		;IDLE
		sta $1800		;free all lines
		ldx $1800
-
		cpx $1800		;did a bit arrive? (bit flip in data in, atn is dropped in same go in first bit)
		beq *-3
		ldx $1800		;load register
		cpx #$04		;push bit into carry
		ror			;shift in
		bcc -			;do until our counter bit ends up in carry

		eor #$ff		;invert, as bits arrive inverted
		rts

.upload
		;receive end_address of code being uploaded
		;maximum is $0110 - $04ae, buffers at $0500,$0600,$0700 can be used
		;if original code is restored, be sure that get_byte is at the same address

		;receive dest
		jsr .get_byte
		tay
		jsr .get_byte
		sta .dest+1

		;endaddr
		jsr .get_byte
		sta .cmp1
		jsr .get_byte
		sta .cmp2
-
		jsr .get_byte
		sta (.dest),y			;lowbyte of dest is always 0
		iny
		bne +
		inc .dest+1
+
		cpy .cmp1
		bne -
		lda .dest+1
		cmp .cmp2
		bne -
		jmp .drivecode_
		;register $1800 is not $00 or at least $08 here, this sucks, on second turn it is then $00

;emit warnings only once
!if * > $0500 { !serious "Drivecode is ", * - $0500, " bytes too big!" }
!ifdef .drivecode_end {
	!if $0500 - * < $80 {
		!warn $0500 - *, " bytes remaining for drivecode."
	}
}

;all code form here on might be overwriten upon init of loader, as dir is loaded and both nibble buffers are populated
.drivecode_setup
		;clear wanted list and copy over gcr2ser table
		ldx #$17
-
		lda #$ff
		sta .wanted,x
		lda .gcr2ser_orig,x
		sta .gcr2ser + 8,x
		lda .ht_copy,x
		sta .halftrack_code,x
		lda .mt_copy,x
		sta .motor_code,x
		lda .ic_copy,x
		sta .idle_code,x
		dex
		bpl -

		;generate table for index << 3
		ldx #$09
-
		txa
		asl
		asl
		asl
		tay
		lda .gcr_dec_lo,x
		bmi +			;skip gaps ($ff) to not destroy the code placed there
		sta .gcr_dec_lo_shf,y
		lda .gcr_dec_hi,x
		sta .gcr_dec_hi_shf,y
		lda .gcr2ser,x
		sta .gcr2ser,y
+
		inx
		tya
		bne -

		ldx #$0f
-
		lda .bin2ser_orig,x
		sta .bin2ser,x
		dex
		bpl -
		lda #18
		sta .track
!if BITFIRE_CONFIG_IN_ORDER = 1 {
		lda #$00
		sta .windex
}
		jsr .read_dir
		jmp .idle
!if * > $0500 { !serious "Driveinit end > $0500, so it will overwrite itself upon reading dir!" }

;code snippets that will be slapped into ZP to save some mem
.ht_copy
!pseudopc .halftrack_code {
.halftrack
		tya
		eor $1c00
		sec
		rol
		and #3
		eor $1c00
		sta $1c00
		bit $1c05
		bmi *-3
		rts
}
.mt_copy
!pseudopc .motor_code {
.motor_on
		lda $1c00
		ora #.VIA2_LED_ON
.motor_on_no_led
		ora #.VIA2_MOTOR_ON
		sta $1c00
		rts
}

.ic_copy
!pseudopc .idle_code {
		lda $1c00
!if BITFIRE_CONFIG_MOTOR_ALWAYS_ON = 0 {
		and #(.VIA2_LED_OFF & .VIA2_MOTOR_OFF)
} else {
		and #.VIA2_LED_OFF
}
		sta $1c00
		jmp .get_byte
}

.gcr2ser_orig
		!byte $ff,$0e,$0f,$07,$ff,$0a,$0b,$03
		!byte $ff,$ff,$0d,$05,$ff,$00,$09,$01
		!byte $ff,$06,$0c,$04,$ff,$02,$08,$ff
.bin2ser_orig
		!byte %1111, %0111, %1101, %0101, %1011, %0011, %1001, %0001
		!byte %1110, %0110, %1100, %0100, %1010, %0010, %1000, %0000

}

.drivecode_end
.drivecode_size = .drivecode_end - .drivecode_start
